package com.github.difflib;

/**
 * Test constants
 *
 * @author simon.mittermueller@gmail.com
 *
 */
public final class TestConstants {

    public static final String BASE_FOLDER_RESOURCES = "target/test-classes/";
    /**
     * The base folder containing the test files.
     */
    public static final String MOCK_FOLDER = BASE_FOLDER_RESOURCES + "/mocks/";

    private TestConstants() {
    }
}
